<?php

namespace FiloBlu\Flow\Model\Actions\From;

/*
 TODO: Handle configurable as input and set categories on all childs
 */

/*
    The value on input can be:
    - a single number (58)
    - a single string (parka)
    - a path of strings (woman/jacket/parka) <-- search case insensitive based on the category admin value name
    - a path of numbers (2/5/9)
    - multiple single values (2||9) (woman||parka) (woman/jacket/parka||woman/jacket/coats) <-- multi values must be separated by double pipes '||'

    Backend configuration:
    - enable/disable import
    - enable/disable assign product to full category path
    - enable/disable assign categories only on configurables
    - enable/disable merge the new categories on the old ones

    Additional info:
    - Empty input value will be skipped (no categories will be deleted)
    - 'path of strings' need to start with the category root (Default category/SS18/Woman/Outerwear/PARKA|TRENCH)
    - 'paths' need to exist (after conversion in ids) inside the table 'catalog_category_entity'
*/

use Exception;
use FiloBlu\Flow\Helper\LoggerProvider;
use FiloBlu\Flow\Helper\Product;
use Magento\Catalog\Model\CategoryFactory;
use Magento\Catalog\Model\Product\Type;
use Magento\Catalog\Model\ProductFactory;
use Magento\Catalog\Model\ResourceModel\Category\CollectionFactory;
use Magento\ConfigurableProduct\Model\ResourceModel\Product\Type\Configurable;
use Monolog\Logger;

/**
 * Class Addtocategory
 * @package FiloBlu\Flow\Model\Actions\From
 */
class Addtocategory
{

    /**
     * @var string
     */
    CONST ENTITY_TYPE_CODE = 'catalog_product';

    /**
     * The untouchable master root category (level 0)
     *
     * @var string
     */
    protected $_master_root_category_id = '1';

    /**
     * @var array
     */
    protected $_categories_textstrpath_to_id = [];
    /**
     * @var array
     */
    protected $_categories_textidpath_to_id = [];
    /**
     * @var array
     */
    protected $_categories_full_ids_path_array_from_id = [];

    /**
     * All the root categories (level 1)
     *
     * @var array
     */
    protected $_root_category_id = [];
    /**
     * @var \FiloBlu\Flow\Helper\Monolog\Logger|Logger
     */
    protected $_logger;
    /**
     * @var null
     */
    protected $_namesPaths = null;
    /**
     * @var Product
     */
    protected $helperProduct;
    /**
     * @var Configurable
     */
    protected $configurableProduct;
    /**
     * @var CategoryFactory
     */
    protected $categoryFactory;
    /**
     * @var ProductFactory
     */
    protected $productFactory;
    /**
     * @var CollectionFactory
     */
    protected $categoryCollectionFactory;

    /**
     * Addtocategory constructor.
     * @param Product $helper_product
     * @param LoggerProvider $loggerProvider
     * @param CategoryFactory $categoryFactory
     * @param ProductFactory $productFactory
     * @param Configurable $configurableProduct
     * @param CollectionFactory $categoryCollectionFactory
     */
    public function __construct(
        Product $helper_product,
        LoggerProvider $loggerProvider,
        CategoryFactory $categoryFactory,
        ProductFactory $productFactory,
        Configurable $configurableProduct,
        CollectionFactory $categoryCollectionFactory
    )
    {
        $this->_logger = $loggerProvider->getLogger();
        $this->helperProduct = $helper_product;
        $this->categoryFactory = $categoryFactory;
        $this->productFactory = $productFactory;
        $this->configurableProduct = $configurableProduct;
        $this->categoryCollectionFactory = $categoryCollectionFactory;
        $this->_initCategories();
    }

    /**
     * @return $this
     */
    protected function _initCategories()
    {
        $collection = $this->categoryCollectionFactory->create();
        $collection->setStore(0);
        $collection->addNameToResult();

        foreach ($collection as $category) {
            $structure = explode('/', $category->getPath());

            // Saving the root categories ids
            if ($category->getLevel() === 1) {
                $this->_root_category_id[] = $category->getId();
            }

            // Removing the category ID 1, I don't need that
            if (($key = array_search('1', $structure, false)) !== false) {
                unset($structure[$key]);
            }

            // Generating path without the first master root category
            $path_no_master_root = implode('/', $structure);

            // Removing empty arrays, I don't need those
            if (count($structure) === 0) {
                continue;
            }

            $path_with_names = [];
            foreach ($structure as $single_category_id) {
                $path_with_names[] = strtolower($collection->getItemById($single_category_id)->getName());
            }
            $full_string = implode('/', $path_with_names);

            $leaf_id = end($structure);

            $this->_categories_textstrpath_to_id[$full_string] = $leaf_id;
            $this->_categories_textidpath_to_id[$path_no_master_root] = $leaf_id;
            $this->_categories_full_ids_path_array_from_id[$leaf_id] = $structure;
        }

        return $this;
    }

    /**
     * @param $product
     * @param $value
     * @param $storeId
     * @throws Exception
     */
    public function process($product, $value, $storeId)
    {

        if (empty($value)) {
            throw new Exception("Invalid value supplied '{$value}'");
        }

        if (empty($this->_categories_full_ids_path_array_from_id) || empty($this->_categories_textstrpath_to_id) || empty($this->_categories_textidpath_to_id) || empty($this->_root_category_id)) {
            $this->_initCategories();
        }

        $value = trim($value);

        $originalValue = $value;


        // gestione piu' categorie
        // $_new_cat_ids will be always an array
        $_new_cat_ids = explode('||', $originalValue);

        // security
        if (count($_new_cat_ids) === 0) {
            throw new Exception("Invalid value supplied '{$value}'");
        }

        // $cat_id can be:
        // single number (58)
        // single string (parka)
        // path of strings (woman/jacket/parka)
        // path of number (2/5/9)
        $categories_ids_array_to_add = [];
        foreach ($_new_cat_ids as $_cat_id) {

            // This can return also multiple categories
            // Depends on the imput and backend config
            // Will return always an array
            $category_ids = $this->findTheCategoryId($_cat_id);

            if (count($category_ids) > 0) {
                foreach ($category_ids as $single_category) {
                    $categories_ids_array_to_add[] = $single_category;
                }
            }

            if (!$category_ids || count($category_ids) === 0) {
                $this->_logger->info("Category '{$_cat_id}'' does not exist - SKIPPED ");
            }
        }

        // Nothing to do
        if (count($categories_ids_array_to_add) === 0) {
            return;
        }

        $merged_category = false;

        $_parent_inherit_categories_of_simple = $this->helperProduct->getParentInheritCategoriesOfSimple();

        // First of all I need to add the categories on the configurable
        // Because the categories on the configurables will be always setted
        // If this is a configurable sku then this condition will be skipped
        if ($product->getTypeId() === Type::DEFAULT_TYPE) {
            $parentIds = $this->configurableProduct->getParentIdsByChild($product->getId());
            if ($_parent_inherit_categories_of_simple && count($parentIds) > 0) {
                foreach ($parentIds as $parentId) {
                    $parent_product = $this->productFactory->create()->load($parentId);
                    // Salvataggio delle categorie
                    $merged_category = $this->getCalculatedCategoriesList($parent_product, $categories_ids_array_to_add);
                    $parent_product->setCategoryIds($merged_category)->save();
                }
            }
        }
        if ($product->getTypeId() === 'configurable') {
            $merged_category = $this->getCalculatedCategoriesList($product, $categories_ids_array_to_add);
            $product->setCategoryIds($merged_category)->save();
        }

        $is_simple_without_parent = false;
        if (!$merged_category) {
            // At this point $product is a simple without parents
            $is_simple_without_parent = true;
            $merged_category = $this->getCalculatedCategoriesList($product, $categories_ids_array_to_add);
        }

        // If is a simple without parent then place the categories and end the procedure here
        if ($is_simple_without_parent) {
            $product->setCategoryIds($merged_category);
            $product->save();
            return;
        }

        //Now that the parents have the categories do I need to set them also on the simple?
        $_only_configurable_category = $this->helperProduct->getOnlyConfigurableCategoryFlag();

        // If the simple is child of a parent and it should have the categories
        if (!$is_simple_without_parent && !$_only_configurable_category) {
            $product->setCategoryIds($merged_category);
            $product->save();
            return;
        }
    }

    /**  $anything can be:
     * > single number (9)
     * > single string (parka)
     * > path of strings (default category/woman/jacket/parka)
     * > path of numbers (2/5/9)
     * @param $anything
     * @return array|bool|mixed
     */
    public function findTheCategoryId($anything)
    {
        $anything = strtolower(trim($anything));

        $_categories_textstrpath_to_id = $this->_categories_textstrpath_to_id;
        $_categories_textidpath_to_id = $this->_categories_textidpath_to_id;
        $_categories_full_ids_path_array_from_id = $this->_categories_full_ids_path_array_from_id;

        // single number (58)
        if (is_numeric($anything) && $anything > 0) {
            return isset($_categories_full_ids_path_array_from_id[$anything]) && $this->assignFullPath() ? $_categories_full_ids_path_array_from_id[$anything] : [$anything];
        }

        // single string (parka)
        // path of strings (woman/jacket/parka)
        if (isset($_categories_textstrpath_to_id[$anything])) {
            if ($this->assignFullPath()) {
                return $_categories_full_ids_path_array_from_id[$_categories_textstrpath_to_id[$anything]];
            }

            return [$_categories_textstrpath_to_id[$anything]];
        }

        // path of numbers (2/5/9)
        if (isset($_categories_textidpath_to_id[$anything])) {
            if ($this->assignFullPath()) {
                return $_categories_full_ids_path_array_from_id[$_categories_textidpath_to_id[$anything]];
            }

            return [$_categories_textidpath_to_id[$anything]];
        }

        // not found
        return false;
    }

    /**
     * @return mixed
     */
    public function assignFullPath()
    {
        return $this->helperProduct->assignFullTree();
    }

    /**
     * This function will get the $product categories and handle if those need to be merged with the new ones or not
     * Returning the array of categories to be saved
     * @param $product
     * @param $categories_ids_array_to_add
     * @return array
     */
    private function getCalculatedCategoriesList($product, $categories_ids_array_to_add)
    {
        $_category_update = $this->mergeCategories();
        $old_categories = $product->getCategoryIds();

        if ($_category_update && count($old_categories)) {
            $merged_category = array_merge($old_categories, $categories_ids_array_to_add);
        } else {
            $merged_category = $categories_ids_array_to_add;
        }

        $merged_category = array_unique($merged_category);
        $merged_category = array_diff($merged_category, [$this->_master_root_category_id]); // Tolgo la master root category
        if (!$this->assignRootCategories() && count($this->_root_category_id) > 0) {
            $merged_category = array_diff($merged_category, $this->_root_category_id); // Tolgo le root categories
        }
        return $merged_category;
    }

    /**
     * @return mixed
     */
    public function mergeCategories()
    {
        return $this->helperProduct->mergeCategoriesFlag();
    }

    /**
     * @return mixed
     */
    public function assignRootCategories()
    {
        return $this->helperProduct->assignRootCategories();
    }

}